/***********************************
 * Decrypt files
 * 10/9/2002
 * The Flying Pig
 ***********************************/

#define _CD

//#define _RISC_OS
#define _MS_WIN

#define VERSION  "1.00 (25 Sep 2002)"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <miracl.h>
#include <time.h>
#include <ctype.h>

#if defined _RISC_OS
#include <kernel.h>
#include <sys/swis.h>
#endif
#if defined _MS_WIN
#include <windows.h>
#endif

// Filename defines
#if defined _RISC_OS
#if defined _CD
#define     szFILE_ROOT              "CDFS::DLJ_CARD.$"
#else
#define     szFILE_ROOT              "ATAFS::Cosmas.$.Files.CDCard"
#endif
#define     szFILE_KEEP_DEFAULT      "Encrypted.Files"
#define     szFILE_LIST_ENCRYPT      "Encrypted.Data.EncList/dat"
#define     szFILE_LIST_DECRYPT      "Encrypted.Data.DecList/dat"
#define     szFILE_DOCS_DEFAULT      "<DecryptSave$Dir>"
#define     szDIRECTORY_SEPARATOR    "."
#define     szEXTENSION_SEPARATOR    "/"
#define     szROOT_SEPARATOR         "$"
#elif defined _MS_WIN
#if defined _CD
#define     szFILE_ROOT              "D:"
#else
#define     szFILE_ROOT              "C:\\crypt\\MIRACL"
#endif
#define     szFILE_KEEP_DEFAULT      "Encrypted\\Files"
#define     szFILE_DOCS_DEFAULT      "C:\\Documents and Settings\\All Users\\Desktop"
#define     szFILE_LIST_ENCRYPT      "Encrypted\\Data\\EncList.dat"
#define     szFILE_LIST_DECRYPT      "Encrypted\\Data\\DecList.dat"
#define     szDIRECTORY_SEPARATOR    "\\"
#define     szEXTENSION_SEPARATOR    "."
#define     szROOT_SEPARATOR         ":"
#else
#if defined _CD
#define     szFILE_ROOT              "D:/crypt/MIRACL"
#else
#define     szFILE_ROOT              "D:"
#endif
#define     szFILE_KEEP_DEFAULT      "Encrypted/Files"
#define     szFILE_DOCS_DEFAULT      "C:/tmp"
#define     szFILE_LIST_ENCRYPT      "Encrypted/Data/EncList.dat"
#define     szFILE_LIST_DECRYPT      "Encrypted/Data/DecList.dat"
#define     szDIRECTORY_SEPARATOR    "/"
#define     szEXTENSION_SEPARATOR    "."
#define     szROOT_SEPARATOR         ":"
#endif

#define     uMAX_FILE_LEN            (255)

#define     nBLOCK_SIZE              (16 * 16)

typedef struct _FileHeader
{
  int  nSize;
  int  nFileType;
} FileHeader;

/* This is fixed */
#define NB 4

typedef struct _EncryptFilelist
{
//  char  cKeyPress;
//  char  szLeafName[31];
  char  szPassword[32];
} EncryptFilelist;

typedef struct _DecryptFilelist
{
  char  cKeyPress;
  char  szLeafName[31];
} DecryptFilelist;

char szFileRoot[uMAX_FILE_LEN] = szFILE_ROOT;
char szFileLoadDefault[uMAX_FILE_LEN] = szFILE_KEEP_DEFAULT;
char szFileSaveDefault[uMAX_FILE_LEN] = szFILE_DOCS_DEFAULT;
char szFileListEncrypt[uMAX_FILE_LEN] = szFILE_LIST_ENCRYPT;
char szFileListDecrypt[uMAX_FILE_LEN] = szFILE_LIST_DECRYPT;
char szDirSep[2] = szDIRECTORY_SEPARATOR;
char szExtSep[2] = szEXTENSION_SEPARATOR;
char szRootSep[2] = szROOT_SEPARATOR;

#if defined _MS_WIN
HANDLE      ghStdin;
#endif

/* Use elliptic curve of the form y^2=x^3-3x+B */

/* NIST p224 bit elliptic curve prime 2#224-2#96+1 */

char * ecp = "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000001";

/* Elliptic curve parameter B */

char * ecb = "B4050A850C04B3ABF54132565044B0B7D7BFD8BA270B39432355FFB4";

/* Elliptic curve - point of prime order (x,y) */

char * ecx = "F7D13920ECA3006B6FA69B37A1178888E7E7DAB446ACF673331DD48A";
char * ecy = "E312F42445ED3C782899948B8379FAE620819FA084EF617C185FA6DC";

static miracl *         gMip;

static int              gPause;

void DecryptDecrypt ( char * pData, long int nSize, char * pszKey );
void DecryptObtainKey (char * pszKey, char cKeyPress);
void KeyStoreEncrypt ( char * pData, long int nSize, char * pszKey );
void KeyStoreDecrypt ( char * pData, long int nSize, char * pszKey );
void EncryptEncrypt ( char * pData, long int nSize, char * pszKey );
void EncryptDecrypt ( char * pData, long int nSize, char * pszKey );

int DecryptMain(int argc, char * * argv);
int EncryptMain(int argc, char * * argv);
int KeyStoreMain(int argc, char * * argv);

char * FileSepConv (char * const pszNameOut, char const * pszNameIn);
void PressKeyFinish (void);
void ShowSyntax (void);
void GSTrans(char * const szTranslated, char const * const szIn, int nLen);

////////////////////////////////////////////////////////////////////////
// DECRYPT
////////////////////////////////////////////////////////////////////////

int main(int argc, char * * argv)
{
  int    nCount;

  gPause = 1;

  for (nCount = 0; nCount < argc; nCount++)
  {
    if (strcmp (argv[nCount], "-c") == 0)
    {
      gPause = 0;
    }
  }

  if (argc <= 1)
  {
    return DecryptMain(argc, argv);
  }
  else
  {
    if (argv[1][0] == '-')
    {
      switch (argv[1][1])
      {
        case 'D':
          return DecryptMain(argc, argv);
          break;
        case 'E':
          return EncryptMain(argc, argv);
          break;
        case 'K':
          return KeyStoreMain(argc, argv);
          break;
        default:
        case 'h':
          ShowSyntax();
          PressKeyFinish();
          break;
      }
    }
    else
    {
      ShowSyntax();
      PressKeyFinish();
    }
  }
  return 0;
}

////////////////////////////////////////////////////////////////////////
// DECRYPT
////////////////////////////////////////////////////////////////////////


int DecryptMain(int argc, char * * argv)
{
    char             szFileIn[256];
    char             szFileOut[256];
    char             szKey[17];
    FILE *           psFileIn;
    FILE *           psFileOut;
    long int         nFileInSize;
    long int         nFileOutSize;
    char *           pData;
//    int              nRead;
    FileHeader *     pFileHeader;
    int              nReadSize;
    int              nWriteSize;
    int              nFileType;

    big              bKey;
    char *           pPointer;
    int              nCount;
    int              nSearchForRoot;
    int              nHeight;
    aes              sAes;
    char             cIv[16];
    unsigned long    uSeed;
    FILE *           psFileDecFL;
    int              nFLSize;
    DecryptFilelist *    asDecFL;
    char             cKeyPress = 0;
    int              nDecFile;
#if defined _RISC_OS
    _kernel_swi_regs sRegisters;
    int              nCarry;
#endif
#if defined _MS_WIN
    DWORD            nRead;
#endif

#ifndef MR_NOFULLWIDTH
    gMip = mirsys(100, 0);
#else
    gMip = mirsys(100, MAXBASE);
#endif
    bKey = mirvar(0);
    time((time_t *) & uSeed);
    irand(uSeed);
    gMip->IOBASE = 36;

    gPause = 1;

    nSearchForRoot = 1;
    for (nCount = 0; nCount < argc; nCount++)
    {
      if (strcmp (argv[nCount], "-r") == 0)
      {
        nSearchForRoot = 0;
      }
    }

    // Set up the root
    if ((argc > 0) && nSearchForRoot)
    {
      GSTrans(szFileIn, argv[0], 256);
      nCount = strlen(szFileIn);
      psFileDecFL = NULL;
      nHeight = 3;
      while ((nCount > 0) && (nHeight > 0))
      {
        while ((nCount > 0) && (szFileIn[nCount] != szDirSep[0]))
        {
          nCount--;
        }
        if (szFileIn[nCount] == szDirSep[0])
        {
          szFileIn[nCount + 1] = 0;
          sprintf(szFileOut, "%s%s", szFileIn, szFileListDecrypt);
          psFileDecFL = fopen(szFileOut, "rb");
          if (psFileDecFL)
          {
            // We've found the file we want
            strcpy(szFileRoot, szFileIn);
            szFileRoot[nCount] = 0;
            fclose(psFileDecFL);
            nHeight = 0;
          }
          if (szFileIn[nCount - 1] == szRootSep[0])
          {
            nHeight = 0;
          }
          nHeight--;
        }
        nCount--;
      }
    }

    printf("Decrypt version %s\n", VERSION);
    printf("The Flying Pig\n");

    for (nCount = 0; nCount < argc; nCount++)
    {
      if ((argv[nCount][0] == '-') && (strlen(argv[nCount]) == 2))
      {
        switch (argv[nCount][1])
        {
          case 'D':
               break;
          case 'c':
               gPause = 0;
               break;
          case 'r':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileRoot, "%s", argv[nCount + 1]);
               break;
          case 's':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileSaveDefault, "%s", argv[nCount + 1]);
               break;
          case 'e':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileListEncrypt, "%s", argv[nCount + 1]);
               break;
          case 'd':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileListEncrypt, "%s", argv[nCount + 1]);
               break;
          case 'f':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileLoadDefault, "%s", argv[nCount + 1]);
               break;
          case 'k':
               if ((nCount == (argc - 1)) || (strlen (argv[nCount + 1]) != 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               cKeyPress = argv[nCount + 1][0];
               break;
          default:
               ShowSyntax ();
               return 0;
               break;
        }
      }
    }


    // Set up all of the paths
    sprintf(szFileIn, "%s%s%s", szFileRoot, szDirSep, szFileListEncrypt);
    GSTrans(szFileListEncrypt, szFileIn, 256);
    sprintf(szFileIn, "%s%s%s", szFileRoot, szDirSep, szFileListDecrypt);
    GSTrans(szFileListDecrypt, szFileIn, 256);
    sprintf(szFileIn, "%s%s%s%s", szFileRoot, szDirSep, szFileLoadDefault, szDirSep);
    GSTrans(szFileLoadDefault, szFileIn, 256);

    printf("Root set to %s\n\n", szFileRoot);

    // Open the file (read binary)
    psFileDecFL = fopen(szFileListDecrypt, "rb");
    if (psFileDecFL)
    {
      // Move to the end of the file
      fseek(psFileDecFL, 0, SEEK_END);
      // Find out where we are (and hence how large the file is)
      nFileInSize = ftell(psFileDecFL);
    }
    else
    {
      nFileInSize = 0;
    }
    // Allocate a suitable amount of memory
    asDecFL = (DecryptFilelist *)malloc(nFileInSize);
    // The number of elements in the array we've just created
    nFLSize = (nFileInSize / sizeof(DecryptFilelist));

    // Load the whole file in to memory
    if (nFileInSize > 0)
    {
      fseek(psFileDecFL, 0, SEEK_SET);
      fread(asDecFL, sizeof(DecryptFilelist), nFLSize, psFileDecFL);
    }
    // Close the file
    if (psFileDecFL)
    {
      fclose(psFileDecFL);
    }

#if defined _RISC_OS
    // Convert the filenames
    for (nCount = 0; nCount < nFLSize; nCount++)
    {
      FileSepConv (asDecFL[nCount].szLeafName, asDecFL[nCount].szLeafName);
    }
#endif

    // Files now loaded - we can continue

    if (cKeyPress == 0)
    {
      // Print the file contents on to the screen
      printf("File list: \n\n");
      for (nCount = 0; nCount < nFLSize; nCount += 2)
      {
        printf("%c. %-32s", asDecFL[nCount].cKeyPress, asDecFL[nCount].szLeafName);
        if ((nCount + 1) < nFLSize)
        {
          printf("     %c. %-s", asDecFL[nCount + 1].cKeyPress, asDecFL[nCount + 1].szLeafName);
        }
        printf("\n");
      }
      printf("\n");


      printf("Press the key corresponding to the file you wish to decrypt:\n");

#if defined _RISC_OS
      _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
      cKeyPress = sRegisters.r[0];
      printf("%c\n", cKeyPress);
#elif defined _MS_WIN
      ghStdin = GetStdHandle(STD_INPUT_HANDLE);
      SetConsoleMode(ghStdin, (ENABLE_PROCESSED_INPUT));
      ReadConsoleA(ghStdin, & cKeyPress, 1, & nRead, NULL);
      SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
      printf("%c\n", cKeyPress);
#else
      scanf(" %c%*c", & cKeyPress);
#endif
    }
    // Find out which file is intended for decryption (using the file list)
    // Search through the list of keys until we find the first match
    for (nDecFile = 0; (nDecFile < nFLSize) && (cKeyPress != asDecFL[nDecFile].cKeyPress); nDecFile++)
    {
    }
    if (cKeyPress != asDecFL[nDecFile].cKeyPress)
    {
      printf("This option does not exist!\n");
      PressKeyFinish();
      return 0;
    }
    else
    {
      printf("You have chosen the file '%s'\n", asDecFL[nDecFile].szLeafName);
    }

    // Set up the file paths
    sprintf(szFileIn, "%s%s", szFileLoadDefault, asDecFL[nDecFile].szLeafName);

    printf("Enter a directory to save the file to: \n");
    gets(szFileOut);
    // Remove any postfixed whitespace
    nCount = 0;
    while ((nCount < (int)sizeof(szFileOut)) && (szFileOut[nCount] != 0))
    {
      nCount++;
    }
    nCount--;
    while ((nCount > 0) && isspace(szFileOut[nCount]))
    {
      szFileOut[nCount] = 0;
      nCount--;
    }

    // Concatonate the directory
    if (strncmp(szFileOut, "", 2) == 0)
    {
      GSTrans(szFileSaveDefault, szFileSaveDefault, 256);
      printf("%s\n", szFileSaveDefault);
      sprintf(szFileOut, "%s%s%s", szFileSaveDefault, szDirSep, asDecFL[nDecFile].szLeafName);
    }
    else
    {
      GSTrans(szFileOut, szFileOut, 256);
      sprintf(szFileOut, "%s%s%s", szFileOut, szDirSep, asDecFL[nDecFile].szLeafName);
    }

    // Clear the key
    for (nCount = 0; nCount < (int)sizeof(szKey); nCount++)
    {
      szKey[nCount] = 0;
    }

    DecryptObtainKey (szKey, cKeyPress);

    printf("\nSaving file %s\n", szFileOut);

    gMip->IOBASE = 36;

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes,MR_CBC,16,szKey,cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return 0;
    }

    // Open the file (read binary)
    psFileIn = fopen(szFileIn, "rb");

    if (!psFileIn)
    {
      printf("Failed to open input file\n");
      PressKeyFinish();
      return 0;
    }
    // Open the file (write binary)
    psFileOut = fopen(szFileOut, "wb");
    if (!psFileOut)
    {
      printf("Could not open output file\n");
      PressKeyFinish();
      return 0;
    }
    // Move to the end of the file
    fseek(psFileIn, 0, SEEK_END);
    // Find out where we are (and hence how large the file is)
    nFileInSize = ftell(psFileIn);

    // Allocate a suitable amount of memory
    pData = (char *)calloc(1, nBLOCK_SIZE);

    // Load the start of the file in to memory
    fseek(psFileIn, 0, SEEK_SET);
    if (nFileInSize < nBLOCK_SIZE)
    {
      nReadSize = nFileInSize;
      nWriteSize = nFileInSize - sizeof(FileHeader);
    }
    else
    {
      nReadSize = nBLOCK_SIZE;
      nWriteSize = nBLOCK_SIZE - sizeof(FileHeader);
    }
    fread(pData, 1, nReadSize, psFileIn);

    // Decrypt this chunk of data
    pPointer = pData;
    while (pPointer < (pData + nReadSize))
    {
      // Decrypt 16 bytes worth
      aes_decrypt( & sAes, pPointer);
      pPointer += 16;
    }

    // Get the file's header information
    pFileHeader = (FileHeader *)pData;
    nFileOutSize = pFileHeader->nSize;
    nFileType = pFileHeader->nFileType;

    // Do a rudimentary check to prevent a gigantic file being saved
    // if the key is wrong
    if (nFileOutSize > nFileInSize)
    {
      printf("Key appears to be incorrect - unable to decrypt file\n");
      aes_end(& sAes);
      // Close the file
      fclose(psFileIn);
      fclose(psFileOut);

      // Free up the allocated memory
      free (asDecFL);
      PressKeyFinish();
      return 0;
    }

    if (nFileOutSize < nWriteSize)
    {
      nWriteSize = nFileOutSize;
    }

    // Save this chunk out
    fwrite(pData + sizeof(FileHeader), 1, nWriteSize, psFileOut);

    nFileInSize -= nReadSize;
    nFileOutSize -= nWriteSize;

    // If theres any file left, encrypt and save this too
    while (nFileOutSize > 0)
    {
      if (nFileInSize < nBLOCK_SIZE)
      {
        nReadSize = nFileInSize;
        nWriteSize = nFileInSize;
      }
      else
      {
        nReadSize = nBLOCK_SIZE;
        nWriteSize = nBLOCK_SIZE;
      }
      fread(pData, 1, nReadSize, psFileIn);

      // Decrypt this chunk of data
      pPointer = pData;
      while (pPointer < (pData + nWriteSize))
      {
        // Decrypt 16 bytes worth
        aes_decrypt(&sAes, pPointer);
        pPointer += 16;
      }

      if (nFileOutSize < nWriteSize)
      {
        nWriteSize = nFileOutSize;
      }

      nFileInSize -= nReadSize;
      nFileOutSize -= nWriteSize;

      // Save this chunk out
      fwrite(pData, 1, nWriteSize, psFileOut);
    }

    aes_end(& sAes);
    // Close the file
    fclose(psFileIn);
    fclose(psFileOut);

    // Set the filetype
#if defined _RISC_OS
    sRegisters.r[0] = 18;
    sRegisters.r[1] = (int)szFileOut;
    sRegisters.r[2] = nFileType;
    _kernel_swi(OS_File, & sRegisters, & sRegisters);
#endif

    // Free up the allocated memory
    free (asDecFL);

    printf("Decryption complete\n");
    PressKeyFinish();
    return 0;
}

void DecryptDecrypt ( char * pData, long int nSize, char * pszKey )
{
    char          szKey[17];
    char *        pPointer;
    int           nCount;
    aes           sAes;
    char          cIv[16];
//    unsigned long uSeed;
    miracl *      mip;

#ifndef MR_NOFULLWIDTH
    mip = mirsys(100, 0);
#else
    mip = mirsys(100, MAXBASE);
#endif

    // Clear the key
    for (nCount = 0; nCount < (int)sizeof(szKey); nCount++)
    {
      szKey[nCount] = 0;
    }

    strncpy (szKey, pszKey, sizeof(szKey));

    mip->IOBASE = 36;

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes, MR_CBC, 16, szKey, cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return;
    }

    aes_reset(& sAes, MR_CBC, cIv);

    pPointer = pData;

    while (pPointer < (pData + nSize))
    {
      // Decrypt 16 bytes worth
      aes_decrypt(& sAes, pPointer);
      pPointer += 16;
    }

    aes_end(& sAes);
}

void DecryptObtainKey (char * pszKey, char cKeyPress)
{
    char             szKey[255];
    char             szTemp[255];
    char             szTempN1[255];
    char             szTempN2[255];
    char             szEncryptKey[17];
    char             szFileKey[17];
    int              nPassSend;
    int              nPassReceive;
//    unsigned long    uSeed;
    epoint *         g, * ea, * eb;
    big              a, b, p;
    big              bPassSend, bPassReceive, bKey, bCurveX, bCurveY;
    big              bFileKey;
    int              nCount;
//#ifndef MR_NOFULLWIDTH
//    gMip = mirsys(100, 0);
//#else
//    gMip = mirsys(100, MAXBASE);
//#endif
    a = mirvar(0);
    b = mirvar(0);
    p = mirvar(0);
    bFileKey = mirvar(0);
    bPassSend = mirvar(0);
    bPassReceive = mirvar(0);
    bKey = mirvar(0);
    bCurveX = mirvar(0);
    bCurveY = mirvar(0);
//    time((time_t *) & uSeed);
//    irand(uSeed);   /* change parameter for different values */

/*
   Elliptic Curve public key swapping.
   Curve is y^2=x^3+Ax+B mod p, where A=-3, B and p as above
   "Primitive root" is the point (x,y) above, which is of large prime order q.
   In this case actually
   q= FFFFFFFFFFFFFFFFFFFFFFFFFFFF16A2E0B8F03E13DD29455C5C2A3D

*/

    convert(-3, a);
    gMip->IOBASE = 16;
    cinstr(b, ecb);
    cinstr(p, ecp);
    ecurve_init(a, b, p, MR_PROJECTIVE);

    g = epoint_init();
    cinstr(bCurveX, ecx);
    cinstr(bCurveY, ecy);
    gMip->IOBASE = 36;
    epoint_set(bCurveX, bCurveY, 0, g);
    ea = epoint_init();
    eb = epoint_init();
    epoint_copy(g, ea);
    epoint_copy(g, eb);

    bigdig(160, 2, a);
    ecurve_mult(a, ea, ea);
    nPassSend = epoint_get(ea, bPassSend, bPassSend);
    /* <ia,pa> is compressed form of public key */

    cotstr(bPassSend, szTempN1);
    printf("\nPassword to send:\n%d %d %s\n", (cKeyPress - 32), nPassSend, szTempN1);

    printf("\nEnter returned password: \n");
    scanf(" %d %s %s%*c", & nPassReceive, szKey, szTemp);

    szKey[255] = 0;
    szTemp[255] = 0;
    cinstr(bPassReceive, szKey);

    epoint_set(bPassReceive, bPassReceive, nPassReceive, eb);
    /* decompress eb */

    ecurve_mult(a, eb, eb);
    // Sets up bKey as the final encryption key
    // which is the same on both machines
    epoint_get(eb, bKey, bKey);

    // Convert the encrypted file key entered by the user
    gMip->IOBASE = 36;
    cinstr(bFileKey, szTemp);
    gMip->IOBASE = 256;
    szFileKey[16] = 0;
    for (nCount = 0; nCount < (int)sizeof(szTemp); nCount++){szTemp[nCount] = 0;}
    for (nCount = 0; nCount < (int)sizeof(szFileKey); nCount++){szFileKey[nCount] = 0;}
    cotstr(bFileKey, szTemp);
    strncpy(szFileKey, szTemp, sizeof(szFileKey));

//    // Convert this in to a big number so that we can decrypt it
//    cotnum(bFileKey, stdout);
//    gMip->IOBASE = 256;
//    big_to_bytes(16, bFileKey, szFileKey, FALSE);


    // Use this key to decrypt the file key
    gMip->IOBASE = 36;
    for (nCount = 0; nCount < (int)sizeof(szTemp); nCount++){szTemp[nCount] = 0;}
    for (nCount = 0; nCount < (int)sizeof(szEncryptKey); nCount++){szEncryptKey[nCount] = 0;}
    cotstr(bKey, szTemp);
    strncpy(szEncryptKey, szTemp, sizeof(szEncryptKey));
    szEncryptKey[16] = 0;

    // Decrypt the key we're going to use to decrypt the file with
    DecryptDecrypt(szFileKey, 16, szEncryptKey);

    // Convert the encrypted output in to something we can display on screen
    gMip->IOBASE = 256;
    cinstr(bFileKey, szFileKey);

    // Place the various keys into strings so that we can print them
    gMip->IOBASE = 36;

    cotstr(bFileKey, szTempN2);

    strncpy(pszKey, szTempN2, 16);

    epoint_free(g);
    epoint_free(ea);
    epoint_free(eb);

    return;
}

////////////////////////////////////////////////////////////////////////
// ENCRYPT
////////////////////////////////////////////////////////////////////////

/***********************************
 * Encrypt files
 * 10/9/2002
 * The Flying Pig
 ***********************************/

int EncryptMain(int argc, char * * argv)
{
    char             szFileIn[255];
    char             szFileOut[255];
    char             szFileOutSave[255];
    char             szKey[17];
    char             szPassword[17];
    FILE *           psFileIn;
    FILE *           psFileOut;
    FILE *           psFileEncFL;
    FILE *           psFileDecFL;
    long int         nFileInSize;
    long int         nFileOutSize;
    char *           pData;
//    int              nRead;
    int              nCorrect;
    int              nCheckChar;
    FileHeader *     pFileHeader;
    int              n32ReadSize;
    int              n32WriteSize;
    big              bKey;
    char *           pPointer;
    int              nCount;
    int              nSearchForRoot;
    int              nHeight;
    aes              sAes;
    char             cIv[16];
    unsigned long    uSeed;
    EncryptFilelist *    asEncFL;
    int             nFLSize;
    DecryptFilelist *    asDecFL;
    char            cKeyPress = 0;
    int             nDecFile;
    int             nReplaceFile;
#if defined _RISC_OS
    _kernel_swi_regs sRegisters;
    int              nCarry;
#endif
#if defined _MS_WIN
    DWORD            nRead;
#endif

#ifndef MR_NOFULLWIDTH
    gMip = mirsys(100,0);
#else
    gMip = mirsys(100,MAXBASE);
#endif
    bKey = mirvar(0);
    time((time_t *) & uSeed);
    irand(uSeed);
    gMip->IOBASE = 36;

    gPause = 1;

    nSearchForRoot = 1;
    for (nCount = 0; nCount < argc; nCount++)
    {
      if (strcmp (argv[nCount], "-r") == 0)
      {
        nSearchForRoot = 0;
      }
    }

    // Set up the root
    if ((argc > 0) && nSearchForRoot)
    {
      GSTrans(szFileIn, argv[0], 256);
      nCount = strlen(szFileIn);
      psFileDecFL = NULL;
      nHeight = 3;
      while ((nCount > 0) && (nHeight > 0))
      {
        while ((nCount > 0) && (szFileIn[nCount] != szDirSep[0]))
        {
          nCount--;
        }
        if (szFileIn[nCount] == szDirSep[0])
        {
          szFileIn[nCount + 1] = 0;
          sprintf(szFileOut, "%s%s", szFileIn, szFileListDecrypt);
          psFileDecFL = fopen(szFileOut, "rb");
          if (psFileDecFL)
          {
            // We've found the file we want
            strcpy(szFileRoot, szFileIn);
            szFileRoot[nCount] = 0;
            fclose(psFileDecFL);
            nHeight = 0;
          }
          if (szFileIn[nCount - 1] == szRootSep[0])
          {
            nHeight = 0;
          }
          nHeight--;
        }
        nCount--;
      }
    }

    printf("Encrypt version %s\n", VERSION);
    printf("The Flying Pig\n");

    for (nCount = 0; nCount < argc; nCount++)
    {
      if ((argv[nCount][0] == '-') && (strlen(argv[nCount]) == 2))
      {
        switch (argv[nCount][1])
        {
          case 'E':
               break;
          case 'c':
               gPause = 0;
               break;
          case 'r':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileRoot, "%s", argv[nCount + 1]);
               break;
/*
          case 's':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileSaveDefault, "%s", argv[nCount + 1]);
               break;
*/
          case 'e':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileListEncrypt, "%s", argv[nCount + 1]);
               break;
          case 'd':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileListEncrypt, "%s", argv[nCount + 1]);
               break;
          case 'f':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileLoadDefault, "%s", argv[nCount + 1]);
               break;
/*
          case 'k':
               if ((nCount == (argc - 1)) || (strlen (argv[nCount + 1]) != 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               cKeyPress = argv[nCount + 1][0];
               break;
*/
          default:
               ShowSyntax ();
               return 0;
               break;
        }
      }
    }

    // Set up all of the paths
    sprintf(szFileIn, "%s%s%s", szFileRoot, szDirSep, szFileListEncrypt);
    GSTrans(szFileListEncrypt, szFileIn, 256);
    sprintf(szFileIn, "%s%s%s", szFileRoot, szDirSep, szFileListDecrypt);
    GSTrans(szFileListDecrypt, szFileIn, 256);
    sprintf(szFileIn, "%s%s%s%s", szFileRoot, szDirSep, szFileLoadDefault, szDirSep);
    GSTrans(szFileLoadDefault, szFileIn, 256);

    printf("Root set to %s\n\n", szFileRoot);

    // Clear the password completely
    for (nCount = 0; nCount < (int)sizeof(szPassword); nCount++)
    {
      szPassword[nCount] = 0;
    }

    // Get the main password
    printf("Enter main password: \n");

#if defined _RISC_OS
    nCount = 0;
    cKeyPress = '0';
    while ((nCount < (int)sizeof(szPassword)) && (cKeyPress >= '0'))
    {
      _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
      cKeyPress = sRegisters.r[0];
      szPassword[nCount] = cKeyPress;
      nCount++;
    }
    nCount--;
    szPassword[nCount] = 0;
#elif defined _MS_WIN
    ghStdin = GetStdHandle(STD_INPUT_HANDLE);
    SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT));
    scanf(" %s%*c", szPassword);
    SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
#else
    scanf(" %s%*c", szPassword);
#endif

    szPassword[16] = 0;
    // Load in the two filelists
//    printf("Password entered = %s\n", szPassword);
    // Open the file (read binary)
    psFileEncFL = fopen(szFileListEncrypt, "rb");
    if (psFileEncFL)
    {
      // Move to the end of the file
      fseek(psFileEncFL, 0, SEEK_END);
      // Find out where we are (and hence how large the file is)
      nFileInSize = ftell(psFileEncFL);
    }
    else
    {
      nFileInSize = 0;
    }
    // Allocate a suitable amount of memory
    asEncFL = (EncryptFilelist *)malloc(nFileInSize + sizeof(EncryptFilelist));
    // The number of elements in the array we've just created
    nFLSize = (nFileInSize / sizeof(EncryptFilelist)) + 1;

    // Load the whole file in to memory
    if (nFileInSize > 0)
    {
      fseek(psFileEncFL, 0, SEEK_SET);
      fread(asEncFL, sizeof(EncryptFilelist), (nFLSize - 1), psFileEncFL);
      // Decrypt this data
      EncryptDecrypt ( (char *)asEncFL, nFileInSize, szPassword );
    }
    // Close the file
    if (psFileEncFL)
    {
      fclose(psFileEncFL);
    }

    // Parse the decrypted data to see if the password was (probably) correct
    nCorrect = -1;
    for (nCount = 0; nCount < (nFLSize - 1); nCount++)
    {
      for (nCheckChar = 0; nCheckChar < 16; nCheckChar++)
      {
        if (asEncFL[nCount].szPassword[nCheckChar] == 0)
        {
          nCheckChar = 16;
        }
        else
        {
          if ((asEncFL[nCount].szPassword[nCheckChar] < '0')
              || (asEncFL[nCount].szPassword[nCheckChar] > 'Z'))
          {
            nCorrect = 0;
          }
        }
      }
    }
    if (nCorrect == 0)
    {
      printf("Incorrect password - cannot decrypt keys file\n");
      PressKeyFinish();
      return 0;
    }

    // Open the file (read binary)
    psFileDecFL = fopen(szFileListDecrypt, "rb");
    if (psFileDecFL)
    {
      // Move to the end of the file
      fseek(psFileDecFL, 0, SEEK_END);
      // Find out where we are (and hence how large the file is)
      nFileInSize = ftell(psFileDecFL);
    }
    else
    {
      nFileInSize = 0;
    }
    // Allocate a suitable amount of memory
    asDecFL = (DecryptFilelist *)malloc(nFileInSize + sizeof(DecryptFilelist));
    // The number of elements in the array we've just created
    if ((int)(nFLSize) != (int)((nFileInSize / sizeof(DecryptFilelist)) + 1))
    {
      printf("ERROR: File list sizes are different\n");
      PressKeyFinish();
      return 0;
    }

    // Load the whole file in to memory
    if (nFileInSize > 0)
    {
      fseek(psFileDecFL, 0, SEEK_SET);
      fread(asDecFL, sizeof(DecryptFilelist), (nFLSize - 1), psFileDecFL);
    }
    // Close the file
    if (psFileDecFL)
    {
      fclose(psFileDecFL);
    }

#if defined _RISC_OS
    // Convert the filenames
    for (nCount = 0; nCount < nFLSize; nCount++)
    {
      FileSepConv (asDecFL[nCount].szLeafName, asDecFL[nCount].szLeafName);
    }
#endif

    // Files now loaded - we can continue

    // Print the file contents on to the screen
    printf("\nFile list (for reference): \n\n");
    for (nCount = 1; nCount < nFLSize; nCount += 2)
    {
      printf("%c. %-32s", asDecFL[nCount - 1].cKeyPress, asDecFL[nCount - 1].szLeafName);
      if ((nCount + 1) < nFLSize)
      {
        printf("     %c. %-s", asDecFL[nCount].cKeyPress, asDecFL[nCount].szLeafName);
      }
      printf("\n");
    }
    printf("\n");

    // Clear the password completely
    for (nCount = 0; nCount < (int)sizeof(szKey); nCount++)
    {
      szKey[nCount] = 0;
    }

    printf("Enter the pathname of the file to encrypt: \n");

    gets(szFileIn);
    // Remove any postfixed whitespace
    nCount = 0;
    while ((nCount < (int)sizeof(szFileIn)) && (szFileIn[nCount] != 0))
    {
      nCount++;
    }
    nCount--;
    while ((nCount > 0) && isspace(szFileIn[nCount]))
    {
      szFileIn[nCount] = 0;
      nCount--;
    }

    GSTrans(szFileIn, szFileIn, 256);

    printf("Enter the leafname of the file to output: \n");
    gets(szFileOut);
    // Remove any postfixed whitespace
    nCount = 0;
    while ((nCount < (int)sizeof(szFileOut)) && (szFileOut[nCount] != 0))
    {
      nCount++;
    }
    nCount--;
    while ((nCount > 0) && isspace(szFileOut[nCount]))
    {
      szFileOut[nCount] = 0;
      nCount--;
    }

    printf("Enter the keypress which the file will be referred to by: \n");
#if defined _RISC_OS
    _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
    cKeyPress = sRegisters.r[0];
    printf("%c\n", cKeyPress);
#elif defined _MS_WIN
    SetConsoleMode(ghStdin, (ENABLE_PROCESSED_INPUT));
    ReadConsoleA(ghStdin, & cKeyPress, 1, & nRead, NULL);
    SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
    printf("%c\n", cKeyPress);
#else
    scanf(" %c%*c", & cKeyPress);
#endif

    sprintf(szFileOutSave, "%s%s", szFileLoadDefault, szFileOut);
    printf("Input filename: %s\n", szFileIn);
    printf("Output leafname: %s\n", szFileOut);
    printf("Output filename: %s\n", szFileOutSave);
    printf("Keypress: %c\n", cKeyPress);

    // Discover if this is a repeated key
    // Search through the list of keys until we find a match
    for (nDecFile = 0; (nDecFile < (nFLSize - 1)) && (cKeyPress != asDecFL[nDecFile].cKeyPress); nDecFile++)
    {
    }
    if (cKeyPress != asDecFL[nDecFile].cKeyPress)
    {
      printf("Enter key ('r' for random): \n");
      scanf(" %s%*c", szKey);
      nReplaceFile = 0;
    }
    else
    {
      printf("This key already exists for file '%s'\n", asDecFL[nDecFile].szLeafName);
      printf("This file will be replaced by the new file\n");
      printf("Enter key ('p' for previous key, 'r' for random): \n");
      scanf(" %s%*c", szKey);
      nReplaceFile = -1;
      nFLSize--;
    }

//    printf("Key: %s\n", szKey);

    gMip->IOBASE = 36;

    // Set up the key to encrypt the file with
    if (strncmp(szKey, "r", 2) == 0)
    {
      // Choose a random key
      printf("Using random key\n");
      bigdig(16, 36, bKey);
    }
    else
    {
      if ((nReplaceFile) && (strncmp(szKey, "p", 2) == 0))
      {
        printf("Using previous key\n");
        cinstr(bKey, asEncFL[nDecFile].szPassword);
      }
      else
      {
        printf("Using given key\n");
        cinstr(bKey, szKey);
      }
    }

    cotstr(bKey, szKey);
//    printf("Key used is the following: %s \n", szKey);

    // Set up the details in the file lists
    strncpy(asEncFL[nDecFile].szPassword, szKey, 32);
    asEncFL[nDecFile].szPassword[31] = 0;

//    strncpy(asEncFL[nFLSize - 1].szLeafName, szFileOut, 32);
//    asEncFL[nFLSize - 1].szLeafName[30] = 0;

    strncpy(asDecFL[nDecFile].szLeafName, szFileOut, 31);
    asDecFL[nDecFile].szLeafName[30] = 0;

//    asEncFL[nDecFile].cKeyPress = cKeyPress;
    asDecFL[nDecFile].cKeyPress = cKeyPress;

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes, MR_CBC, 16, szKey, cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return 0;
    }

    // Open the file (read binary)
    psFileIn = fopen(szFileIn, "rb");
    // Open the file (write binary)
    psFileOut = fopen(szFileOutSave, "wb");
    // Move to the end of the file
    fseek(psFileIn, 0, SEEK_END);
    // Find out where we are (and hence how large the file is)
    nFileInSize = ftell(psFileIn);

    // We have to load the file into an area with size a multiple of 16 bytes
    // We're also going to add an extra word at the beginning
    // to store the original file size in
    nFileOutSize = 16 * ((nFileInSize + sizeof(FileHeader) + 15) / 16);


    // Allocate a suitable amount of memory
    pData = (char *)calloc(1, nBLOCK_SIZE);

    // Set up the file header
    pFileHeader = (FileHeader *)pData;
    pFileHeader->nSize = nFileInSize;
#if defined _RISC_OS
    sRegisters.r[0] = 23;
    sRegisters.r[1] = (int)szFileIn;
    _kernel_swi(OS_File, & sRegisters, & sRegisters);
    pFileHeader->nFileType = sRegisters.r[6];
#endif

    // Load the start of the file in to memory
    fseek(psFileIn, 0, SEEK_SET);
    if (nFileInSize < (nBLOCK_SIZE - (int)sizeof(FileHeader)))
    {
      n32ReadSize = nFileInSize;
      n32WriteSize = 16 * ((nFileInSize + sizeof(FileHeader) + 15) / 16);
    }
    else
    {
      n32ReadSize = (nBLOCK_SIZE - sizeof(FileHeader));
      n32WriteSize = nBLOCK_SIZE;
    }
    fread(pData + sizeof(FileHeader), 1, n32ReadSize, psFileIn);

    // Encrypt this chunk of data
    pPointer = pData;
    while (pPointer < (pData + n32WriteSize))
    {
      // Encrypt 16 bytes worth
      aes_encrypt(&sAes, pPointer);
      pPointer += 16;
    }

    nFileInSize -= n32ReadSize;

    // Save this chunk out
    fwrite(pData, 1, n32WriteSize, psFileOut);

    // If theres any file left, encrypt and save this too
    while (nFileInSize > 0)
    {
      if (nFileInSize < nBLOCK_SIZE)
      {
        n32ReadSize = nFileInSize;
        n32WriteSize = 16 * ((nFileInSize + 15) / 16);
      }
      else
      {
        n32ReadSize = nBLOCK_SIZE;
        n32WriteSize = nBLOCK_SIZE;
      }
      fread(pData, 1, n32ReadSize, psFileIn);

      // Encrypt this chunk of data
      pPointer = pData;
      while (pPointer < (pData + n32WriteSize))
      {
        // Encrypt 16 bytes worth
        aes_encrypt(&sAes, pPointer);
        pPointer += 16;
      }

      nFileInSize -= n32ReadSize;

      // Save this chunk out
      fwrite(pData, 1, n32WriteSize, psFileOut);
    }

    aes_end(& sAes);
    // Close the file
    fclose(psFileIn);
    fclose(psFileOut);

    printf("File encrypted and saved\n");

#if defined _RISC_OS
    // Convert the filenames
    for (nCount = 0; nCount < nFLSize; nCount++)
    {
      FileSepConv (asDecFL[nCount].szLeafName, asDecFL[nCount].szLeafName);
    }
#endif

    // Encrypt the filelist data
    EncryptEncrypt ( (char *)asEncFL, nFLSize * sizeof(EncryptFilelist), szPassword );
    // Save the new filelists
    // Open the file (write binary)
    // this will overwrite the previous version
    psFileEncFL = fopen(szFileListEncrypt, "wb");
    // Save the whole file from memory
    fseek(psFileEncFL, 0, SEEK_SET);
    fwrite(asEncFL, sizeof(EncryptFilelist), nFLSize, psFileEncFL);
    // Close the file
//    fflush(psFileEncFL);
    fclose(psFileEncFL);
    // Open the file (write binary)
    // this will overwrite the previous version
    psFileDecFL = fopen(szFileListDecrypt, "wb");
    // Save the whole file from memory
    fseek(psFileDecFL, 0, SEEK_SET);
    fwrite(asDecFL, sizeof(DecryptFilelist), nFLSize, psFileDecFL);
    // Close the file
//    fflush(psFileDecFL);
    fclose(psFileDecFL);

    printf("New file list data saved\n");
    // Free up the allocated memory
    free (asEncFL);
    free (asDecFL);

    PressKeyFinish();
    return 0;
}

void EncryptEncrypt ( char * pData, long int nSize, char * pszKey )
{
    char            szKey[17];
    big             bKey;
    char *          pPointer;
    int             nCount;
    aes             sAes;
    char            cIv[16];
//    unsigned long   uSeed;

//#ifndef MR_NOFULLWIDTH
//    gMip = mirsys(100, 0);
//#else
//    gMip = mirsys(100, MAXBASE);
//#endif
    bKey = mirvar(0);
//    time((time_t *) & uSeed);
//    irand(uSeed);

    // Clear the password completely
    for (nCount = 0; nCount < (int)sizeof(szKey); nCount++)
    {
      szKey[nCount] = 0;
    }

    gMip->IOBASE = 36;

    if (strncmp(pszKey, "r", 2) == 0)
    {
      // Choose a random key
      printf("Using random key\n");
      bigdig(16, 36, bKey);
    }
    else
    {
      cinstr(bKey, pszKey);
    }

    cotstr(bKey, szKey);

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes, MR_CBC, 16, szKey, cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return;
    }

    pPointer = pData;

    while (pPointer < (pData + nSize))
    {
      // Encrypt 16 bytes worth
      aes_encrypt(&sAes, pPointer);
      pPointer += 16;
    }

    aes_end(& sAes);
}

void EncryptDecrypt ( char * pData, long int nSize, char * pszKey )
{
    char          szKey[17];
    char *        pPointer;
    int           nCount;
    aes           sAes;
    char          cIv[16];
//    unsigned long uSeed;

//#ifndef MR_NOFULLWIDTH
//    gMip = mirsys(100, 0);
//#else
//    gMip = mirsys(100, MAXBASE);
//#endif

    // Clear the key
    for (nCount = 0; nCount < (int)sizeof(szKey); nCount++)
    {
      szKey[nCount] = 0;
    }

    strncpy (szKey, pszKey, sizeof(szKey));

    gMip->IOBASE = 36;

//    printf("Key used is the following: %s \n", szKey);

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes, MR_CBC, 16, szKey, cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return;
    }

    aes_reset(& sAes,MR_CBC,cIv);

    pPointer = pData;

    while (pPointer < (pData + nSize))
    {
      // Decrypt 16 bytes worth
      aes_decrypt(& sAes, pPointer);
      pPointer += 16;
    }
    aes_end(& sAes);
}

////////////////////////////////////////////////////////////////////////
// KEYSTORE
////////////////////////////////////////////////////////////////////////

/***********************************
 * Store and transmit encrypted keys
 * 10/9/2002
 * The Flying Pig
 ***********************************/

int KeyStoreMain(int argc, char * * argv)
{
    char            szFileIn[256];
    char            szFileOut[256];
    char            szKey[48];
    char            szTemp[255];
    char            szTempN1[255];
    char            szTempN2[255];
    char            szEncryptKey[17];
    char            szFileKey[17];
    char            szPassword[17];
    int             nPassSend;
    int             nPassReceive;
    unsigned long   uSeed;
    epoint *        g, * ea, * eb;
    big             a, b, p;
    big             bPassSend, bPassReceive, bKey, bCurveX, bCurveY;
    big             bFileKey;

    FILE *          psFileEncFL;
    FILE *          psFileDecFL;
    long int        nFileInSize;
//    long int        nFileOutSize;
    EncryptFilelist *    asEncFL;
    int             nFLSize;
    DecryptFilelist *    asDecFL;
    char            cKeyPress = 0;
    int             nKeyPress;
    int             nDecFile;
    int             nCount;
    int             nSearchForRoot;
    int             nHeight;
    int             nCheckChar;
    int             nCorrect;
#if defined _RISC_OS
    _kernel_swi_regs sRegisters;
    int             nCarry;
#endif
#if defined _MS_WIN
//    DWORD           nRead;
#endif

#ifndef MR_NOFULLWIDTH
    gMip = mirsys(100, 0);
#else
    gMip = mirsys(100, MAXBASE);
#endif
    a = mirvar(0);
    b = mirvar(0);
    p = mirvar(0);
    bFileKey = mirvar(0);
    bPassSend = mirvar(0);
    bPassReceive = mirvar(0);
    bKey = mirvar(0);
    bCurveX = mirvar(0);
    bCurveY = mirvar(0);
    time((time_t *) & uSeed);
    irand(uSeed);   /* change parameter for different values */
    gMip->IOBASE = 36;

    gPause = 1;

    nSearchForRoot = 1;
    for (nCount = 0; nCount < argc; nCount++)
    {
      if (strcmp (argv[nCount], "-r") == 0)
      {
        nSearchForRoot = 0;
      }
    }

    // Set up the root
    if ((argc > 0) && nSearchForRoot)
    {
      GSTrans(szFileIn, argv[0], 256);
      nCount = strlen(szFileIn);
      psFileDecFL = NULL;
      nHeight = 3;
      while ((nCount > 0) && (nHeight > 0))
      {
        while ((nCount > 0) && (szFileIn[nCount] != szDirSep[0]))
        {
          nCount--;
        }
        if (szFileIn[nCount] == szDirSep[0])
        {
          szFileIn[nCount + 1] = 0;
          sprintf(szFileOut, "%s%s", szFileIn, szFileListDecrypt);
          psFileDecFL = fopen(szFileOut, "rb");
          if (psFileDecFL)
          {
            // We've found the file we want
            strcpy(szFileRoot, szFileIn);
            szFileRoot[nCount] = 0;
            fclose(psFileDecFL);
            nHeight = 0;
          }
          if (szFileIn[nCount - 1] == szRootSep[0])
          {
            nHeight = 0;
          }
          nHeight--;
        }
        nCount--;
      }
    }

    printf("KeyStore version %s\n", VERSION);
    printf("The Flying Pig\n");

    for (nCount = 0; nCount < argc; nCount++)
    {
      if ((argv[nCount][0] == '-') && (strlen(argv[nCount]) == 2))
      {
        switch (argv[nCount][1])
        {
          case 'K':
               break;
          case 'c':
               gPause = 0;
               break;
          case 'r':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileRoot, "%s", argv[nCount + 1]);
               break;
/*
          case 's':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileSaveDefault, "%s", argv[nCount + 1]);
               break;
*/
          case 'e':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileListEncrypt, "%s", argv[nCount + 1]);
               break;
          case 'd':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileListEncrypt, "%s", argv[nCount + 1]);
               break;
/*
          case 'f':
               if (nCount == (argc - 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               sprintf(szFileLoadDefault, "%s", argv[nCount + 1]);
               break;
*/
/*
          case 'k':
               if ((nCount == (argc - 1)) || (strlen (argv[nCount + 1]) != 1))
               {
                 ShowSyntax ();
                 return 0;
               }
               cKeyPress = argv[nCount + 1][0];
               break;
*/
          default:
               ShowSyntax ();
               return 0;
               break;
        }
      }
    }

    // Set up all of the paths
    sprintf(szFileIn, "%s%s%s", szFileRoot, szDirSep, szFileListEncrypt);
    GSTrans(szFileListEncrypt, szFileIn, 256);
    sprintf(szFileIn, "%s%s%s", szFileRoot, szDirSep, szFileListDecrypt);
    GSTrans(szFileListDecrypt, szFileIn, 256);
//    sprintf(szFileIn, "%s%s%s%s", szFileRoot, szDirSep, szFileLoadDefault, szDirSep);
//    GSTrans(szFileLoadDefault, szFileIn, 256);

    printf("Root set to %s\n\n", szFileRoot);

    // Get the main password
    printf("Enter main password: \n");

#if defined _RISC_OS
    nCount = 0;
    cKeyPress = '0';
    while ((nCount < (int)sizeof(szPassword)) && (cKeyPress >= '0'))
    {
      _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
      cKeyPress = sRegisters.r[0];
      szPassword[nCount] = cKeyPress;
      nCount++;
    }
    nCount--;
    szPassword[nCount] = 0;
#elif defined _MS_WIN
    ghStdin = GetStdHandle(STD_INPUT_HANDLE);
    SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_PROCESSED_INPUT));
    scanf(" %s%*c", szPassword);
    SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
#else
    scanf(" %s%*c", szPassword);
#endif


    szPassword[16] = 0;
    // Load in the two filelists
//    printf("Password entered = %s\n", szPassword);
    // Open the file (read binary)
    psFileEncFL = fopen(szFileListEncrypt, "rb");
    if (psFileEncFL)
    {
      // Move to the end of the file
      fseek(psFileEncFL, 0, SEEK_END);
      // Find out where we are (and hence how large the file is)
      nFileInSize = ftell(psFileEncFL);
    }
    else
    {
      nFileInSize = 0;
    }
    // Allocate a suitable amount of memory
    asEncFL = (EncryptFilelist *)malloc(nFileInSize);
    // The number of elements in the array we've just created
    nFLSize = (nFileInSize / sizeof(EncryptFilelist));

    // Load the whole file in to memory
    if (nFileInSize > 0)
    {
      fseek(psFileEncFL, 0, SEEK_SET);
      fread(asEncFL, sizeof(EncryptFilelist), nFLSize, psFileEncFL);
      // Decrypt this data
      KeyStoreDecrypt ( (char *)asEncFL, nFileInSize, szPassword );
    }
    // Close the file
    if (psFileEncFL)
    {
      fclose(psFileEncFL);
    }

    // Parse the decrypted data to see if the password was (probably) correct
    nCorrect = -1;
    for (nCount = 0; nCount < nFLSize; nCount++)
    {
      for (nCheckChar = 0; nCheckChar < 16; nCheckChar++)
      {
        if (asEncFL[nCount].szPassword[nCheckChar] == 0)
        {
          nCheckChar = 16;
        }
        else
        {
          if ((asEncFL[nCount].szPassword[nCheckChar] < '0')
              || (asEncFL[nCount].szPassword[nCheckChar] > 'Z'))
          {
            nCorrect = 0;
          }
        }
      }
    }
    if (nCorrect == 0)
    {
      printf("Incorrect password - cannot decrypt keys file\n");
      PressKeyFinish();
      return 0;
    }

    // Open the file (read binary)
    psFileDecFL = fopen(szFileListDecrypt, "rb");
    if (psFileDecFL)
    {
      // Move to the end of the file
      fseek(psFileDecFL, 0, SEEK_END);
      // Find out where we are (and hence how large the file is)
      nFileInSize = ftell(psFileDecFL);
    }
    else
    {
      nFileInSize = 0;
    }
    // Allocate a suitable amount of memory
    asDecFL = (DecryptFilelist *)malloc(nFileInSize);
    // The number of elements in the array we've just created
    if ((int)(nFLSize) != (int)(nFileInSize / sizeof(DecryptFilelist)))
    {
      printf("ERROR: File list sizes are different\n");
      PressKeyFinish();
      return 0;
    }

    // Load the whole file in to memory
    if (nFileInSize > 0)
    {
      fseek(psFileDecFL, 0, SEEK_SET);
      fread(asDecFL, sizeof(DecryptFilelist), nFLSize, psFileDecFL);
    }
    // Close the file
    if (psFileDecFL)
    {
      fclose(psFileDecFL);
    }

#if defined _RISC_OS
    // Convert the filenames
    for (nCount = 0; nCount < nFLSize; nCount++)
    {
      FileSepConv (asDecFL[nCount].szLeafName, asDecFL[nCount].szLeafName);
    }
#endif

    // Files now loaded - we can continue

    // Print the file contents on to the screen
    printf("\nFile list (for reference): \n\n");
    for (nCount = 0; nCount < nFLSize; nCount += 2)
    {
      printf("%c. %-32s", asDecFL[nCount].cKeyPress, asDecFL[nCount].szLeafName);
      if ((nCount + 1) < nFLSize)
      {
        printf("     %c. %-s", asDecFL[nCount + 1].cKeyPress, asDecFL[nCount + 1].szLeafName);
      }
      printf("\n");
    }
    printf("\n");

/*
   Now Elliptic Curve version of the above.
   Curve is y^2=x^3+Ax+B mod p, where A=-3, B and p as above
   "Primitive root" is the point (x,y) above, which is of large prime order q.
   In this case actually
   q= FFFFFFFFFFFFFFFFFFFFFFFFFFFF16A2E0B8F03E13DD29455C5C2A3D

*/

    convert(-3, a);
    gMip->IOBASE = 16;
    cinstr(b, ecb);
    cinstr(p, ecp);
    ecurve_init(a, b, p, MR_PROJECTIVE);

    g = epoint_init();
    cinstr(bCurveX, ecx);
    cinstr(bCurveY, ecy);
    gMip->IOBASE = 36;
    epoint_set(bCurveX, bCurveY, 0, g);
    ea = epoint_init();
    eb = epoint_init();
    epoint_copy(g, ea);
    epoint_copy(g, eb);

    bigdig(160, 2, a);
    ecurve_mult(a, ea, ea);
    nPassSend = epoint_get(ea, bPassSend, bPassSend);
    /* <ia,pa> is compressed form of public key */

    printf("Enter returned password: \n");
    scanf("%d %d %s%*c", & nKeyPress, & nPassReceive, szKey);
  cKeyPress = (char)(nKeyPress + 32);

    // Find out which file is intended for decryption (using the file list)
    // Search through the list of keys until we find the first match
    for (nDecFile = 0; (nDecFile < nFLSize) && (cKeyPress != asDecFL[nDecFile].cKeyPress); nDecFile++)
    {
    }
    if (cKeyPress != asDecFL[nDecFile].cKeyPress)
    {
      printf("File could not be found - is password correct?\n");
      PressKeyFinish();
      return 0;
    }
    else
    {
      printf("This key will decrypt the file '%s'\n", asDecFL[nDecFile].szLeafName);
    }

    cinstr(bPassReceive, szKey);

    epoint_set(bPassReceive, bPassReceive, nPassReceive, eb);
    /* decompress eb */

    ecurve_mult(a, eb, eb);
    // Sets up bKey as the final encryption key
    // which is the same on both machines
    epoint_get(eb, bKey, bKey);

    // Use this key to encrypt the file key
    gMip->IOBASE = 36;
    cotstr(bKey, szTemp);

    strncpy(szEncryptKey, szTemp, sizeof(szEncryptKey));
    szEncryptKey[16] = 0;

    // Make a copy of the key to send, so that we can encrypt it
    szFileKey[16] = 0;
    strncpy(szFileKey, asEncFL[nDecFile].szPassword, sizeof(szFileKey));
    // Encrypt the key we're going to send

    KeyStoreEncrypt(szFileKey, 16, szEncryptKey);

//    gMip->IOBASE = 256;
//    szFileKey[16] = 0;
//    cinstr(bFileKey, szFileKey);


    // Convert the encrypted output in to something we can display on screen
    bytes_to_big(16, szFileKey, bFileKey);

    // Place the various keys into strings so that we can print them
    gMip->IOBASE=36;
    cotstr(bPassSend, szTempN1);
    cotstr(bFileKey, szTempN2);

    printf("\nPassword to send:\n%d %s %s\n", nPassSend, szTempN1, szTempN2);

    epoint_free(g);
    epoint_free(ea);
    epoint_free(eb);

    // Free up the allocated memory
    free (asEncFL);
    free (asDecFL);

    PressKeyFinish();
    return 0;
}

void KeyStoreEncrypt ( char * pData, long int nSize, char * pszKey )
{
    char            szKey[17];
    big             bKey;
    char *          pPointer;
    int             nCount;
    aes             sAes;
    char            cIv[16];
//    unsigned long   uSeed;
//    miracl *        gMip;

//#ifndef MR_NOFULLWIDTH
//    gMip = mirsys(100, 0);
//#else
//    gMip = mirsys(100, MAXBASE);
//#endif
    bKey = mirvar(0);
//    time((time_t *) & uSeed);
//    irand(uSeed);

    szKey[16] = 0;
    gMip->IOBASE = 36;

    if (strncmp(pszKey, "r", 2) == 0)
    {
      // Choose a random key
      printf("Using random key\n");
      bigdig(16, 36, bKey);
    }
    else
    {
      cinstr(bKey, pszKey);
    }

    cotstr(bKey, szKey);

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes, MR_CBC, 16, szKey, cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return;
    }

    pPointer = pData;

    while (pPointer < (pData + nSize))
    {
      // Encrypt 16 bytes worth
      aes_encrypt(&sAes, pPointer);
      pPointer += 16;
    }

    aes_end(& sAes);
}

void KeyStoreDecrypt ( char * pData, long int nSize, char * pszKey )
{
    char          szKey[17];
    char *        pPointer;
    int           nCount;
    aes           sAes;
    char          cIv[16];
//    unsigned long uSeed;
//    miracl *      Mip;

//#ifndef MR_NOFULLWIDTH
//    gMip = mirsys(100, 0);
//#else
//    gMip = mirsys(100, MAXBASE);
//#endif

    // Clear the key
    for (nCount = 0; nCount < (int)sizeof(szKey); nCount++)
    {
      szKey[nCount] = 0;
    }

    strncpy (szKey, pszKey, sizeof(szKey));

    gMip->IOBASE = 36;

//    printf("Key used is the following: %s \n", szKey);

    for (nCount = 0; nCount < 16; nCount++)
    {
      cIv[nCount] = nCount;
    }

    if (!aes_init(& sAes, MR_CBC, 16, szKey, cIv))
    {
      printf("Failed to Initialize\n");
      PressKeyFinish();
      return;
    }

    aes_reset(& sAes,MR_CBC,cIv);

    pPointer = pData;

    while (pPointer < (pData + nSize))
    {
      // Decrypt 16 bytes worth
      aes_decrypt(& sAes, pPointer);
      pPointer += 16;
    }
    aes_end(& sAes);
}

////////////////////////////////////////////////////////////////////////
// GENERAL
////////////////////////////////////////////////////////////////////////

void GSTrans(char * const szTranslated, char const * const szIn, int nLen)
{
#if defined _RISC_OS
    _kernel_swi_regs sRegisters;
    int              nCarry;
    char             szTemp[256];

    sRegisters.r[0] = (int)szIn;
    sRegisters.r[1] = (int)szTemp;
    sRegisters.r[2] = nLen;
    _kernel_swi_c(OS_GSTrans, & sRegisters, & sRegisters, & nCarry);
    szTemp[sRegisters.r[2]] = 0;
    strncpy(szTranslated, szTemp, nLen);
#else
    if (szTranslated != szIn)
    {
      strncpy(szTranslated, szIn, nLen);
    }
#endif
}

char * FileSepConv (char * const pszNameOut, char const * pszNameIn)
{
  int                nPos;

  // Swap "/" and "." characters
  nPos = 0;
  while (pszNameIn[nPos] != 0)
  {
    if (pszNameIn[nPos] == szDirSep[0])
    {
      pszNameOut[nPos] = szExtSep[0];
    }
    else
    {
      if (pszNameIn[nPos] == szExtSep[0])
      {
        pszNameOut[nPos] = szDirSep[0];
      }
      else
      {
        pszNameOut[nPos] = pszNameIn[nPos];
      }
    }
    nPos++;
  }
  pszNameOut[nPos] = 0;
  return pszNameOut;
}

void PressKeyFinish (void)
{
#if defined _RISC_OS
    _kernel_swi_regs sRegisters;
    int              nCarry;
#endif
#if defined _MS_WIN
    DWORD            nRead;
    char             cKeyPress;
#endif
    if (gPause == 1)
    {
#if defined _RISC_OS
      printf("\nPress a key to finish\n");
      _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
#elif defined _MS_WIN
      printf("\nPress a key to finish\n");
      SetConsoleMode(ghStdin, (ENABLE_PROCESSED_INPUT));
      ReadConsoleA(ghStdin, & cKeyPress, 1, & nRead, NULL);
      SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
#else
      printf("\nPress ENTER to finish\n");
      gets();
#endif
    }
}

void ShowSyntax (void)
{
  printf("\nThe following arguments may be used:\n");
  printf("  -D     Decrypt files mode\n");
  printf("  -E     Encrypt files mode\n");
  printf("  -K     KeyStore mode\n");
  printf("  -h     Display this help information\n");
  printf("\nAfter which further options may be added:\n");
  printf("  -c     Continue once complete without waiting for a keypress\n");
  printf("  -r     Set the Root path\n");
  printf("  -d     Set the Decryption list filename relative to the root\n");
  printf("  -e     Set the Encryption list filename relative to the root\n");
  printf("  -f     Set the directory containing the encrypted files\n");
  printf("  -s     Set the default save directory\n");
  printf("  -k     Set the keypress for the file to decrypt\n");
  printf("\nFor more information see the help file, or contact David Llewellyn-Jones at llewelld@for.mat.bham.ac.uk\n\n");
}
